"""
CVR Quantiphyse plugin

Author: Martin Craig <martin.craig@nottingham.ac.uk>
Copyright (c) 2021 University of Nottingham, Martin Craig
"""
from .widget import CvrPetCo2Widget
from .process import CvrPetCo2GlmProcess, CvrPetCo2VbProcess

QP_MANIFEST = {
    "widgets" : [CvrPetCo2Widget],
    "processes" : [CvrPetCo2GlmProcess, CvrPetCo2VbProcess],
}



"""
CVR Quantiphyse plugin

Author: Martin Craig <martin.craig@nottingham.ac.uk>
Copyright (c) 2021 University of Nottingham, Martin Craig
"""

from __future__ import division, unicode_literals, absolute_import, print_function

from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, Citation, TitleWidget, RunWidget
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, BoolOption, NumberListOption, TextOption, FileOption, ChoiceOption
from quantiphyse.utils import LogSource

from ._version import __version__

FAB_CITE_TITLE = "Variational Bayesian inference for a non-linear forward model"
FAB_CITE_AUTHOR = "Chappell MA, Groves AR, Whitcher B, Woolrich MW."
FAB_CITE_JOURNAL = "IEEE Transactions on Signal Processing 57(1):223-236, 2009."

class OptionsWidget(QtWidgets.QWidget, LogSource):

    sig_changed = QtCore.Signal()

    def __init__(self, ivm, parent):
        LogSource.__init__(self)
        QtWidgets.QWidget.__init__(self, parent)
        self.ivm = ivm

class AcquisitionOptions(OptionsWidget):

    N_REGRESSORS = 3

    def __init__(self, ivm, parent):
        OptionsWidget.__init__(self, ivm, parent)

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        self._optbox = OptionBox()
        self._optbox.add("<b>Data</b>")
        self._optbox.add("BOLD timeseries data", DataOption(self.ivm), key="data")
        self._optbox.add("ROI", DataOption(self.ivm, rois=True, data=False), key="roi")

        #self._optbox.add("Physiological data (CO<sub>2</sub>/O<sub>2</sub>)", FileOption(plot_btn=True), key="phys-data")
        #self._optbox.add("Sampling frequency (Hz)", NumericOption(minval=0, maxval=1000, default=100, intonly=True), key="samp-rate")
        self._optbox.add("TR for MRI timeseries (s)", NumericOption(minval=0, maxval=5, default=1.0), key="tr")
        self._optbox.add("Baseline period (s)", NumericOption(minval=0, maxval=200, default=60, intonly=True), key="baseline")
        self._optbox.add("MRI timeseries alignment", ChoiceOption(["Automatic", "Manual"]), key="mri-align")
        self._optbox.option("mri-align").sig_changed.connect(self._align_changed)
        self._optbox.add("MRI timeseries start time (s)", NumericOption(minval=0, maxval=1000, default=0), key="data-start-time")
        vbox.addWidget(self._optbox)

        self._optbox_reg = OptionBox()
        self._optbox_reg.add("<b>Regressors</b>")
        for idx in range(self.N_REGRESSORS):
            self._optbox_reg.add("Regressor %i" % (idx+1), ChoiceOption(["Unprocessed CO2", "Preprocessed pETCO2", "Ramp (linear drift)", "Custom"], ["co2", "petco2", "ramp", "custom"]), 
                             checked=True, default=True, key="type_%i" % (idx+1))
            self._optbox_reg.option("type_%i" % (idx+1)).sig_changed.connect(self._regressor_changed)
            self._optbox_reg.add("Data (CO<sub>2</sub>/O<sub>2</sub>)", FileOption(plot_btn=True), key="data_%i" % (idx+1))
            self._optbox_reg.add("Time resolution (s)", NumericOption(minval=0, maxval=10, default=1), key="tr_%i" % (idx+1))

        vbox.addWidget(self._optbox_reg)

        vbox.addStretch(1)
        self._regressor_changed()
        self._align_changed()

    def _regressor_changed(self):
        for idx in range(self.N_REGRESSORS):
            opts = self._optbox_reg.values()
            extras_visible = "type_%i" % (idx+1) in opts and opts["type_%i" % (idx+1)] != "ramp"
            self._optbox_reg.set_visible("data_%i" % (idx+1), extras_visible)
            self._optbox_reg.set_visible("tr_%i" % (idx+1), extras_visible)

    def _add_regressor_options(self, opts):
        regressors = []
        regressor_types = []
        regressor_trs = []
        reg_opts = self._optbox_reg.values()
        for idx in range(self.N_REGRESSORS):
            regressor_type = reg_opts.get("type_%i" % (idx+1), None)
            if regressor_type is not None:
                if regressor_type != "ramp":
                    regressor_types.append(regressor_type)
                    regressors.append(reg_opts["data_%i" % (idx+1)])
                    regressor_trs.append(reg_opts["tr_%i" % (idx+1)])
                else:
                    # FIXME can't mix file regressors with Numpy array, need to write to tmp file
                    regressor_types.append("custom")
                    regressor_trs.append(opts["tr"])
                    regressors.append(np.linspace(0, 1, self.ivm.data[opts["data"]].nvols))
        opts["regressors"] = ",".join(regressors)
        opts["regressor_trs"] = ",".join(["%.3f" % v for v in regressor_trs])
        opts["regressor_types"] = ",".join(regressor_types)

    def _align_changed(self):
        self._optbox.set_visible("data-start-time", self._optbox.option("mri-align").value == "Manual")

    def options(self):
        opts = self._optbox.values()
        opts.pop("mri-align", None)
        self._add_regressor_options(opts)
        return opts

class FabberVbOptions(OptionsWidget):
    def __init__(self, ivm, parent, acq_options):
        OptionsWidget.__init__(self, ivm, parent)
        self.acq_options = acq_options

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        cite = Citation(FAB_CITE_TITLE, FAB_CITE_AUTHOR, FAB_CITE_JOURNAL)
        vbox.addWidget(cite)

        self._optbox = OptionBox()
        self._optbox.add("<b>Model options</b>")
        self._optbox.add("Infer constant signal offset", BoolOption(default=True), key="infer-sig0")
        self._optbox.add("Infer delay", BoolOption(default=True), key="infer-delay")

        #self._optbox.add("<b>Model fitting options</b>")
        #self._optbox.add("Spatial regularization", BoolOption(default=True), key="spatial")
        self._optbox.add("<b>Output options</b>")
        self._optbox.add("Output data name suffix", TextOption(), checked=True, key="output-suffix")

        vbox.addWidget(self._optbox)
        vbox.addWidget(RunWidget(self, save_option=True))
        vbox.addStretch(1)

    def processes(self):
        opts = {
            "model-group" : "cvr",
            "model" : "cvr_petco2",
            "save-mean" : True,
            "save-model-fit" : True,
            "noise" : "white",
            "max-iterations" : 10,
        }
        opts.update(self.acq_options.options())
        opts.update(self._optbox.values())

        # Fabber model requires the physiological data to be preprocessed
        from vaby.data import DataModel
        data_model = DataModel(opts["data"].raw(), mask=opts["mask"].raw())
        from vaby_models_cvr.petco2 import CvrPetCo2Model
        opts["phys_data"] = opts["phys-data"] # FIXME hack
        model = CvrPetCo2Model(data_model, **opts)
        opts["phys-data"] = model.co2_mmHg
        opts.pop("phys_data")

        # Deal with the output suffix if specified
        suffix = opts.pop("output-suffix", "")
        if suffix and suffix[0] != "_":
            suffix = "_" + suffix
        opts["output-rename"] = {
                "mean_cvr" : "cvr%s" % suffix,
                "mean_sig0" : "sig0%s" % suffix,
                "mean_delay" : "delay%s" % suffix,
                "modelfit" : "modelfit%s" % suffix,
        }

        # In spatial mode use sig0 as regularization parameter
        if opts.pop("spatial", False):
            opts["method"] = "spatialvb"
            opts["param-spatial-priors"] = "M+"

        self.debug("Fabber CVR options: %s", opts)
        processes = [
            {"Fabber" : opts},
        ]

        return processes

class VbOptions(OptionsWidget):
    def __init__(self, ivm, parent, acq_options):
        OptionsWidget.__init__(self, ivm, parent)
        self.acq_options = acq_options

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        cite = Citation(FAB_CITE_TITLE, FAB_CITE_AUTHOR, FAB_CITE_JOURNAL)
        vbox.addWidget(cite)

        self._optbox = OptionBox()
        self._optbox.add("<b>Model options</b>")
        self._optbox.add("Infer constant signal offset", BoolOption(default=True), key="infer-sig0")
        self._optbox.add("Infer delay", BoolOption(default=True), key="infer-delay")
        self._optbox.add("Allow negative CVR", BoolOption(default=False), key="allow-neg-cvr")

        self._optbox.add("<b>Model fitting options</b>")
        self._optbox.add("Number of iterations", NumericOption(minval=0, maxval=100, default=10, intonly=True), key="max-iterations")
        #self._optbox.add("Spatial regularization", BoolOption(default=True), key="spatial")

        self._optbox.add("<b>Output options</b>")
        self._optbox.add("Output variance maps", BoolOption(), key="output-var")
        self._optbox.add("Output data name suffix", TextOption(), checked=True, key="output-suffix")

        vbox.addWidget(self._optbox)
        vbox.addWidget(RunWidget(self, save_option=True))
        vbox.addStretch(1)

    def processes(self):
        opts = {}
        opts.update(self.acq_options.options())
        opts.update(self._optbox.values())

        self.debug("CvrPetCo2Vb options: %s", opts)
        processes = [
            {"CvrPetCo2Vb" : opts},
        ]

        return processes

class GlmOptions(OptionsWidget):
    def __init__(self, ivm, parent, acq_options):
        OptionsWidget.__init__(self, ivm, parent)
        self.acq_options = acq_options

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        self._optbox = OptionBox()
        self._optbox.add("<b>Model options</b>")
        self._optbox.add("Delay minimum (s)", NumericOption(minval=-100, maxval=100, default=0), key="delay-min")
        self._optbox.add("Delay maximum (s)", NumericOption(minval=-100, maxval=100, default=0), key="delay-max")
        self._optbox.add("Delay step (s)", NumericOption(minval=-5, maxval=5, default=1), key="delay-step")

        self._optbox.add("<b>Output options</b>")
        self._optbox.add("Output data name suffix", TextOption(), checked=True, key="output-suffix")

        vbox.addWidget(self._optbox)
        vbox.addWidget(RunWidget(self, save_option=True))
        vbox.addStretch(1)

    def processes(self):
        opts = {}
        opts.update(self.acq_options.options())
        opts.update(self._optbox.values())
        self.debug("CvrPetCo2Glm options: %s", opts)
        processes = [
            {"CvrPetCo2Glm" : opts},
        ]

        return processes

class CvrPetCo2Widget(QpWidget):
    """
    CVR modelling of BOLD-MRI with PETCO2
    """
    def __init__(self, **kwargs):
        QpWidget.__init__(self, name="CVR PETCO2", icon="cvr", group="CVR",
                          desc="Cerebrovascular reactivity using BOLD-MRI and PETCO2", **kwargs)
        self.current_tab = 0

    def init_ui(self):
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, help="cvr", subtitle="Cerebrovascular reactivity using BOLD-MRI and PETCO2 v%s" % __version__)
        vbox.addWidget(title)

        self.tabs = QtWidgets.QTabWidget()
        self.tabs.currentChanged.connect(self._tab_changed)
        vbox.addWidget(self.tabs)

        self.acquisition_opts = AcquisitionOptions(self.ivm, parent=self)
        self.tabs.addTab(self.acquisition_opts, "Acquisition Options")
        #self.fabber_opts = FabberVbOptions(self.ivm, self, self.acquisition_opts)
        #self.tabs.addTab(self.fabber_opts, "Bayesian modelling")
        self.vb_opts = VbOptions(self.ivm, self, self.acquisition_opts)
        self.tabs.addTab(self.vb_opts, "Bayesian modelling")
        self.glm_opts = GlmOptions(self.ivm, self, self.acquisition_opts)
        self.tabs.addTab(self.glm_opts, "GLM modelling")

        vbox.addStretch(1)

    def _tab_changed(self):
        tab = self.tabs.currentIndex()
        if tab in (1, 2):
            self.current_tab = tab

    def processes(self):
        # For batch options, return whichever tab was last selected
        # (default to VB options if on acquisition tab)
        if self.current_tab == 1:
            return self.vb_opts.processes()
        elif self.current_tab == 2:
            return self.glm_opts.processes()
        else:
            return []



"""
CVR Quantiphyse plugin - Processes

Author: Martin Craig <martin.craig@nottingham.ac.uk>
Copyright (c) 2021 University of Nottingham, Martin Craig
"""
import io
import logging
import sys
import os

import numpy as np

# Silence Tensorflow random messages
import tensorflow as tf
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
tf.get_logger().setLevel(logging.ERROR)

from quantiphyse.utils import QpException
from quantiphyse.utils.enums import Boundary
from quantiphyse.processes import Process

MAX_LOG_SIZE=100000

def _get_progress_cb(worker_id, queue, n_voxels):
    def _progress(frac):
        queue.put((worker_id, frac * n_voxels))
    return _progress

def _run_glm(worker_id, queue, data, mask, regressors, regressor_types, regressor_trs, tr, baseline, data_start_time, delay_min, delay_max, delay_step):
    try:
        from vaby.data import DataModel
        from vaby_models_cvr.petco2 import CvrPetCo2Model

        options = {
            "regressors" : regressors,
            "regressor_trs" : regressor_trs,
            "regressor_types" : regressor_types,
            "tr" : tr,
            "baseline" : baseline,
            "data_start_time" : data_start_time,
        }

        # Set up log to go to string buffer
        log = io.StringIO()
        handler = logging.StreamHandler(log)
        handler.setFormatter(logging.Formatter('%(levelname)s : %(message)s'))
        logging.getLogger().handlers.clear()
        logging.getLogger().addHandler(handler)
        logging.getLogger().setLevel(logging.INFO)

        data_model = DataModel(data, mask=mask)
        fwd_model = CvrPetCo2Model(data_model, **options)
        glmdata = fwd_model.fit_glm(delay_min=delay_min, delay_max=delay_max, delay_step=delay_step, progress_cb=_get_progress_cb(worker_id, queue, data_model.data_space.size))
        regressor_types = [s.strip() for s in regressor_types.split(",")]
        ret = {}
        for idx, regressor_type in enumerate(regressor_types):
            if regressor_type in ("co2", "petco2"):
                ret["cvr%i" % (idx+1)] = glmdata[idx]
            else:
                ret["beta%i" % (idx+1)] = glmdata[idx]
        ret["delay"] = glmdata[-3]
        ret["sig0"] = glmdata[-2]
        ret["modelfit"] = glmdata[-1]
        for name in list(ret.keys()):
            data = ret[name]
            shape = data_model.data_space.shape
            if data.ndim > 1:
                shape = list(shape) + [data.shape[1]]
            ndata = np.zeros(shape, dtype=np.float32)
            ndata[mask > 0] = data
            ret[name] = ndata
        queue.put((worker_id, data_model.data_space.size))

        ret = (ret, log.getvalue())
        return worker_id, True, ret
    except:
        import traceback
        traceback.print_exc()
        return worker_id, False, sys.exc_info()[1]

class CvrPetCo2GlmProcess(Process):
    """
    CVR-PETCO2 modelling using a GLM
    """
    
    PROCESS_NAME = "CvrPetCo2Glm"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, worker_fn=_run_glm, **kwargs)

    def run(self, options):
        data = self.get_data(options)
        if data.ndim != 4: 
            raise QpException("Data must be 4D for DCE PK modelling")

        roi = self.get_roi(options, data.grid)
    
        self.suffix = options.pop('output-suffix', '')
        if self.suffix != "" and self.suffix[0] != "_": 
            self.suffix = "_" + self.suffix

        regressors = options.pop('regressors', None)
        regressor_trs = options.pop('regressor_trs', None)
        regressor_types = options.pop('regressor_types', None)
        if regressors is None or regressor_types is None or regressor_trs is None:
            raise QpException("Regressors, regressor type and regressor TR options must be given")
        if isinstance(regressors, str):
            regressors = regressors.split()
            new_regressors = []
            for r in regressors:
                if not os.path.isabs(r):
                    r = os.path.join(self.indir, r)
                new_regressors.append(r)
            regressors = ",".join(new_regressors)

        if isinstance(regressor_trs, str):
            try:
                regressor_trs = [float(v) for v in regressor_trs.split(",")]
            except ValueError:
                raise QpException("Regressor TRs should be comma separated list of numbers")
        elif isinstance(regressor_trs, (int, float)):
            regressor_trs = [regressor_trs]
        self.n_regressors = len(regressor_trs)

        tr = options.pop("tr", None)
        if tr is None:
            raise QpException("TR must be given")

        # Non-compulsary options
        baseline = options.pop("baseline", 60)
        data_start_time = options.pop("data-start-time", None)
        delay_min = options.pop("delay-min", 0)
        delay_max = options.pop("delay-max", 0)
        delay_step = options.pop("delay-step", 1)
        
        # Use smallest sub-array of the data which contains all unmasked voxels
        self.grid = data.grid
        self.bb_slices = roi.get_bounding_box()
        self.debug("Using bounding box: %s", self.bb_slices)
        data_bb = data.raw()[tuple(self.bb_slices)]
        mask_bb = roi.raw()[tuple(self.bb_slices)]
        #n_workers = data_bb.shape[0]
        n_workers = 1

        args = [data_bb, mask_bb, regressors, regressor_types, regressor_trs, tr, baseline, data_start_time, delay_min, delay_max, delay_step]
        self.voxels_done = [0] * n_workers
        self.total_voxels = np.count_nonzero(roi.raw())
        self.start_bg(args, n_workers=n_workers)

    def timeout(self, queue):
        if queue.empty(): return
        while not queue.empty():
            worker_id, voxels_done = queue.get()
            self.voxels_done[worker_id] = voxels_done
        progress = float(sum(self.voxels_done)) / self.total_voxels
        self.sig_progress.emit(progress)

    def finished(self, worker_output):
        """
        Add output data to the IVM
        """
        if self.status == Process.SUCCEEDED:
            # Only include log from first process to avoid multiple repetitions
            for data, log in worker_output:
                data_keys = data.keys()
                if log:
                    # If there was a problem the log could be huge and full of
                    # nan messages. So chop it off at some 'reasonable' point
                    self.log(log[:MAX_LOG_SIZE])
                    if len(log) > MAX_LOG_SIZE:
                        self.log("WARNING: Log was too large - truncated at %i chars" % MAX_LOG_SIZE)
                    break
            first = True
            self.data_items = []
            for key in data_keys:
                self.debug(key)
                recombined_data = self.recombine_data([o.get(key, None) for o, l in worker_output])
                name = key + self.suffix
                if key is not None:
                    self.data_items.append(name)
                    if recombined_data.ndim == 2:
                        recombined_data = np.expand_dims(recombined_data, 2)

                    # The processed data was chopped out of the full data set to just include the
                    # ROI - so now we need to put it back into a full size data set which is otherwise
                    # zero.
                    if recombined_data.ndim == 4:
                        shape4d = list(self.grid.shape) + [recombined_data.shape[3],]
                        full_data = np.zeros(shape4d, dtype=np.float32)
                    else:
                        full_data = np.zeros(self.grid.shape, dtype=np.float32)
                    full_data[self.bb_slices] = recombined_data.reshape(full_data[self.bb_slices].shape)
                    self.ivm.add(full_data, grid=self.grid, name=name, make_current=first, roi=False)

                    # Set some view defaults because we know what range these should be in
                    self.ivm.data[name].view.boundary = Boundary.CLAMP
                    if key == "cvr":
                        self.ivm.data[name].view.cmap_range = (0, 1)
                    if key == "delay":
                        self.ivm.data[name].view.cmap_range = (-15, 15)

                    first = False
        else:
            # Include the log of the first failed process
            for out in worker_output:
                if out and isinstance(out, Exception) and hasattr(out, "log") and len(out.log) > 0:
                    self.log(out.log)
                    break

    def output_data_items(self):
        """
        :return: a sequence of data item names that may be output
        """
        data_items = ["cvr", "delay", "sig0"]
        for idx in range(self.n_regressors):
            data_items.append("cvr%i" % (idx+1))
            data_items.append("beta%i" % (idx+1))

        return [key + self.suffix for key in data_items]

def _run_vb(worker_id, queue, data, mask, voxel_sizes, regressors, regressor_types, regressor_trs, tr, infer_sig0, infer_delay, allow_neg_cvr, baseline, data_start_time, spatial, maxits, output_var):
    try:
        from vaby import run

        # Set up log to go to string buffer
        log = io.StringIO()
        options = {
            "method" : "avb",
            "voxel_sizes" : voxel_sizes,
            "regressors" : regressors,
            "regressor_trs" : regressor_trs,
            "regressor_types" : regressor_types,
            "tr" : tr,
            "baseline" : baseline,
            "data_start_time" : data_start_time,
            "infer_sig0" : infer_sig0,
            "infer_delay" : infer_delay,
            "allow_neg_cvr" : allow_neg_cvr,
            "max_iterations" : maxits,
            "save_mean" : True,
            "save_var" : output_var,
            "save_model_fit" : True,
            "log_stream" : log,
            "log_level" : "INFO", # FIXME debugging output
        }
        if spatial:
            options["param_overrides"] = {}
            for param in ("cvr", "delay", "sig0"):
                options["param_overrides"][param] = {"prior_type" : "M"}

        outdict = {}
        _runtime, _state = run(data, "cvr_petco2", mask=mask, outdict=outdict, **options)

        ret = {}
        for name, nii in outdict.items():
            ret[name] = nii.get_fdata()
        ret = (ret, log.getvalue())

        queue.put((worker_id, np.count_nonzero(mask)))
        return worker_id, True, ret
    except:
        import traceback
        traceback.print_exc()
        return worker_id, False, sys.exc_info()[1]

class CvrPetCo2VbProcess(Process):
    """
    CVR-PETCO2 modelling using VB
    """

    PROCESS_NAME = "CvrPetCo2Vb"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, worker_fn=_run_vb, **kwargs)

    def run(self, options):
        data = self.get_data(options)
        if data.ndim != 4:
            raise QpException("Data must be 4D for CVR modelling")

        roi = self.get_roi(options, data.grid)
    
        self.suffix = options.pop('output-suffix', '')
        if self.suffix != "" and self.suffix[0] != "_": 
            self.suffix = "_" + self.suffix

        regressors = options.pop('regressors', None)
        regressor_trs = options.pop('regressor_trs', None)
        regressor_types = options.pop('regressor_types', None)
        if regressors is None or regressor_types is None or regressor_trs is None:
            raise QpException("Regressors, regressor type and regressor TR options must be given")

        if isinstance(regressors, str):
            regressors = regressors.split()
            new_regressors = []
            for r in regressors:
                if not os.path.isabs(r):
                    r = os.path.join(self.indir, r)
                new_regressors.append(r)
            regressors = ",".join(new_regressors)

        if isinstance(regressor_trs, str):
            try:
                regressor_trs = [float(v) for v in regressor_trs.split(",")]
            except ValueError:
                raise QpException("Regressor TRs should be comma separated list of numbers")
        elif isinstance(regressor_trs, (int, float)):
            regressor_trs = [regressor_trs]
        self.n_regressors = len(regressor_trs)

        tr = options.pop("tr", None)
        if tr is None:
            raise QpException("TR must be given")

        # Non-compulsary options
        baseline = options.pop("baseline", 60)
        data_start_time = options.pop("data-start-time", None)
        spatial = options.pop("spatial", False)
        maxits = options.pop("max-iterations", 10)
        self.output_var = options.pop("output-var", False)

        infer_sig0 = options.pop("infer-sig0", True)
        infer_delay = options.pop("infer-delay", True)
        allow_neg_cvr = options.pop("allow-neg-cvr", False)
        
        # Use smallest sub-array of the data which contains all unmasked voxels
        self.grid = data.grid
        self.bb_slices = roi.get_bounding_box()
        self.debug("Using bounding box: %s", self.bb_slices)
        data_bb = data.raw()[tuple(self.bb_slices)]
        mask_bb = roi.raw()[tuple(self.bb_slices)]
        voxel_sizes = self.grid.spacing
        #n_workers = data_bb.shape[0]
        n_workers = 1

        args = [data_bb, mask_bb, voxel_sizes, regressors, regressor_types, regressor_trs, tr, infer_sig0, infer_delay, allow_neg_cvr, baseline, data_start_time, spatial, maxits, self.output_var]
        self.voxels_done = [0] * n_workers
        self.total_voxels = np.count_nonzero(roi.raw())
        self.start_bg(args, n_workers=n_workers)

    def timeout(self, queue):
        if queue.empty(): return
        while not queue.empty():
            worker_id, voxels_done = queue.get()
            self.voxels_done[worker_id] = voxels_done
        progress = float(sum(self.voxels_done)) / self.total_voxels
        self.sig_progress.emit(progress)

    def finished(self, worker_output):
        """
        Add output data to the IVM
        """
        if self.status == Process.SUCCEEDED:
            # Only include log from first process to avoid multiple repetitions
            for data, log in worker_output:
                data_keys = data.keys()
                if log:
                    # If there was a problem the log could be huge and full of 
                    # nan messages. So chop it off at some 'reasonable' point
                    self.log(log[:MAX_LOG_SIZE])
                    if len(log) > MAX_LOG_SIZE:
                        self.log("WARNING: Log was too large - truncated at %i chars" % MAX_LOG_SIZE)
                    break
            first = True
            self.data_items = []
            for key in data_keys:
                self.debug(key)
                recombined_data = self.recombine_data([o.get(key, None) for o, l in worker_output])
                name = key + self.suffix
                if key is not None:
                    self.data_items.append(name)
                    if recombined_data.ndim == 2:
                        recombined_data = np.expand_dims(recombined_data, 2)

                    # The processed data was chopped out of the full data set to just include the
                    # ROI - so now we need to put it back into a full size data set which is otherwise
                    # zero.
                    if recombined_data.ndim == 4:
                        shape4d = list(self.grid.shape) + [recombined_data.shape[3],]
                        full_data = np.zeros(shape4d, dtype=np.float32)
                    else:
                        full_data = np.zeros(self.grid.shape, dtype=np.float32)
                    full_data[self.bb_slices] = recombined_data.reshape(full_data[self.bb_slices].shape)
                    self.ivm.add(full_data, grid=self.grid, name=name, make_current=first, roi=False)

                    # Set some view defaults because we know what range these should be in
                    self.ivm.data[name].view.boundary = Boundary.CLAMP
                    if key == "cvr":
                        self.ivm.data[name].view.cmap_range = (0, 1)
                    if key == "delay":
                        self.ivm.data[name].view.cmap_range = (-15, 15)

                    first = False
        else:
            # Include the log of the first failed process
            for out in worker_output:
                if out and isinstance(out, Exception) and hasattr(out, "log") and len(out.log) > 0:
                    self.log(out.log)
                    break

    def logfile_name(self):
        return "logfile" + self.suffix

    def output_data_items(self):
        """
        :return: a sequence of data item names that may be output
        """
        data_items = ["cvr", "delay", "sig0"]
        for idx in range(self.n_regressors):
            data_items.append("cvr%i" % (idx+1))
            data_items.append("beta%i" % (idx+1))

        if self.output_var:
            for k in list(data_items):
                data_items.append(k + "_var")
        return [key + self.suffix for key in data_items]
